classdef MklExperience < handle
    % USE THIS CLASS TO RUN EXPERIMENTS ON THE OXFORD FLOWER AND CALTECH 101 DATASETS.
    %
    % USAGE EXAMPLE:
    % dtsetFile='/path/to/OxfordFlowers';
    %
    % exp=MklExperience(dtsetDir);
    %
    % classifier='svm'
    % feature=1;
    % C=1;
    % splitNumber=1;
    %
    % [accuracy,model,dec_values,predict_labels,confusion,mca]=exp.runExperiment(classifier,C,splitNumber,feature);
    
    
    properties (Constant)
        LIBSVMPATH='/path/to/libsvm-mat';    %path to the lib-svm solver matlab interface
        SHOGUNPATH='/path/to/shogun2/src/interfaces/matlab_static'; %path to the shogun machine learning toolbox matlab interface
    end
    
    properties
        %GENERAL SETTINGS
        doFromScratch       % if is set to true, the experiment is run from scratch, discarding the precomputed data
        
        %EXPERIMENT PATHS
        datasetPath         % the path where the dataset is stored  (a directory containing another directory called images, where the images are stored)
        kernelsPath         % by default is datasetPath/kernels
        resultsPath         % by default is datasetPath/results
    end
    
    methods
        
        function newExperiment=MklExperience(dataset_path)
            %CONSTRUCTOR OF THE EXPERIMENT
            %
            %PARAMETERS
            % dataset_path is the path where the dataset is stored
            
            kernels_path=sprintf('%s/kernels',dataset_path);
            results_path=sprintf('%s/results',dataset_path);
            
            newExperiment.doFromScratch=false;
            
            newExperiment.datasetPath=dataset_path;
            newExperiment.kernelsPath=kernels_path;
            newExperiment.resultsPath=results_path;
            
            
            %ADDS LIBSVM TO THE PATH
            addpath(MklExperience.LIBSVMPATH);
            
            %ADDS SHOGUN TO THE PATH
            addpath(MklExperience.SHOGUNPATH);
            
            %COMPILES THE AUXILIARY FUNCTION Kbeta
            mex Kbeta.cpp
            
            %CREATE DIRS
            newExperiment.createDirs();
        end
        
        function createDirs(obj)
            %CREATES THE DATASET DIRECTORIES FOR KERNELS AND RESULTS
            
            if ~exist(obj.kernelsPath,'dir')
                mkdir(obj.kernelsPath);
            end
            if ~exist(obj.resultsPath,'dir')
                mkdir(obj.resultsPath);
            end
        end
        
        function [accuracy,model,dec_values,predict_labels,confusion,mca]=runOxfordFlowersExperiment(obj,classifier,C, i, k)
            %RUNS THE EXPERIMENT ON THE OXFORD FLOWERS DATASET AND RETURNS: accuracy, model, decision
            %values, predicted labels, confusion matrix and mean class accuracy
            %
            %PARAMETERS:
            %
            %classifier indicates which classifier to use. It can be either: 'svm',
            %or 'mkl'
            %
            %C is the SVM regularization parameter (higher C means less
            %regularization)
            %
            %i indicates which of the splits is being used
            %
            %k indicates which of the kernels will be used for the
            %experiment (only needed for the single-kernel SVM classifier)
            
            KERNAMES={'Colour','HOG','HSV','Shape','SIFTbdy','SIFTint','Texture','Average','Product'};
            
            
            [Ktr,Kte,ytr,yte]=obj.computeOxfordFlowersKernels(i);
            
            fprintf('Running %s experiment on split: %d.. \n',classifier,i);
            
            if strcmp(classifier,'svm')
                save_file=sprintf('%s/SVM_C%d_%s_N%d.mat',obj.resultsPath,C,KERNAMES{k},i);
                Ktr=Ktr(:,:,k);
                Kte=Kte(:,:,k);
            elseif strcmp(classifier,'mkl')
                Ktr=Ktr(:,:,1:7);
                Kte=Kte(:,:,1:7);
                save_file=sprintf('%s/MKL_C%d_N%d.mat',obj.resultsPath,C,i);
            else
                error('classifier must be either ''svm'', or ''mkl''');
            end
            
            model=MklExperience.trainOva(classifier,Ktr,ytr,C);
            save(save_file,'model');
            [dec_values,predict_labels,accuracy,confusion,mca]=MklExperience.testOva(classifier,Kte,yte,model);
            save(save_file,'model','C','accuracy','dec_values','predict_labels','confusion');
        end
        
        function [accuracy,model,dec_values,predict_labels,confusion,mca]=runCaltech101Experiment(obj,classifier,C, s, k)
            %RUNS THE EXPERIMENT ON THE CALTECH 101 DATASET AND RETURNS: accuracy, model, decision
            %values, predicted labels, confusion matrix and mean class accuracy
            %
            %PARAMETERS:
            %
            %classifier indicates which classifier to use. It can be either: 'svm',
            %or 'mkl'
            %
            %C is the SVM regularization parameter (higher C means less
            %regularization)
            %
            %s is the number of training samples/class. it cam be only 5,
            %15, or 30
            %
            %k indicates which of the kernels will be used for the
            %experiment (only needed for the single-kernel SVM classifier)
            
            KERNAMES={'V1Splus', 'PHOG', 'RegionConvariance','LBP', 'SIFT', 'PHOW', 'SelfSimilarity','Average','Product'};
            
            [Ktr,Kte,ytr,yte]=obj.getCaltech101Kernels(s);
            
            fprintf('Running %s experiment with %d training samples.. \n',classifier,s);
            
            if strcmp(classifier,'svm')
                save_file=sprintf('%s/SVM_C%d_%s_N%d.mat',obj.resultsPath,C,KERNAMES{k},s);
                Ktr=Ktr(:,:,k);
                Kte=Kte(:,:,k);
            elseif strcmp(classifier,'mkl')
                Ktr=Ktr(:,:,1:7);
                Kte=Kte(:,:,1:7);
                save_file=sprintf('%s/MKL_C%d_N%d.mat',obj.resultsPath,C,s);
            else
                error('classifier must be either ''svm'', or ''mkl''');
            end
            
            model=MklExperience.trainOva(classifier,Ktr,ytr,C);
            save(save_file,'model');
            [dec_values,predict_labels,accuracy,confusion,mca]=MklExperience.testOva(classifier,Kte,yte,model);
            save(save_file,'model','C','accuracy','dec_values','predict_labels','confusion');
        end
        
        function [Ktrain, Ktest, Ytrain, Ytest, g]= computeOxfordFlowersKernels(obj,t)
            %COMPUTES THE TRAINING Ktrain AND TESTING Ktest KERNELS
            %
            %PARAMETERS:
            % t indicates which of the splits is being used
            %
            %RETURNS:
            % Ktrain is the 3D training kernel matrix
            % Ktest is the 3D testing kernel matrix
            % Ytrain is the vector of training labels
            % Ytest is the vector of testing labesls
            % g is the vector of parameters of the kernels
            
            feat_file=sprintf('%s/Kernels_N%d.mat',obj.kernelsPath,t);
            
            try
                fprintf('Trying to load kernels for split: %d..\n',t);
                if obj.doFromScratch
                    throw(MException('i:d','DoFromScratch'));
                end
                load(feat_file);
                Ktrain;
                Ktest;
                Ytrain;
                Ytest;
                g;
                fprintf('Loaded Kernels %d.\n',t);
            catch
                fprintf('Failed. Computing kernels for split: %d...\n',t);
                
                % Load the datasplit and distance matrix
                load(sprintf('%s/%s',obj.datasetPath,'datasplits.mat'));
                load(sprintf('%s/%s',obj.datasetPath,'distancematrices17gcfeat06'));
                load(sprintf('%s/%s',obj.datasetPath,'distancematrices17itfeat08.mat'));
                
                Nkern=7;
                eval(['trsplit = trn' num2str(t) ';'])
                eval(['tesplit = tst' num2str(t) ';']);
                
                Ktrain = zeros(680, 680, Nkern);
                Ktest = zeros(340, 680, Nkern);
                
                %COLOUR
                [Ktrain(:,:,1),g(1)]=MklExperience.computeChiSquareKernel(D_colourgc(trsplit, trsplit));
                Ktest(:,:,1)=error('Complete the code to continue..');
                %SHAPE
                [Ktrain(:,:,2),g(2)]=MklExperience.computeChiSquareKernel(D_shapegc(trsplit, trsplit));
                Ktest(:,:,2)=error('Complete the code to continue..');
                %TEXTURE
                [Ktrain(:,:,3),g(3)]=MklExperience.computeChiSquareKernel(D_texturegc(trsplit, trsplit));
                Ktest(:,:,3)=error('Complete the code to continue..');
                %HOG
                [Ktrain(:,:,4),g(4)]=MklExperience.computeChiSquareKernel(D_hog(trsplit, trsplit));
                Ktest(:,:,4)=error('Complete the code to continue..');
                %HSV
                [Ktrain(:,:,5),g(5)]=MklExperience.computeChiSquareKernel(D_hsv(trsplit, trsplit));
                Ktest(:,:,5)=error('Complete the code to continue..');
                %SIFTINT
                [Ktrain(:,:,6),g(6)]=MklExperience.computeChiSquareKernel(D_siftint(trsplit, trsplit));
                Ktest(:,:,6)=error('Complete the code to continue..');
                %SIFTBDY
                [Ktrain(:,:,7),g(7)]=MklExperience.computeChiSquareKernel(D_siftbdy(trsplit, trsplit));
                Ktest(:,:,7)=error('Complete the code to continue..');
                
                %AVERAGE KERNEL
                KtrAvg='Complete the code to continue..\Tip: section 3 in Gehler''s paper';
                KteAvg='Complete the code to continue..\Tip: section 3 in Gehler''s paper';
                
                %PRODUCT KERNEL
                KtrProd='Complete the code to continue..\Tip: section 3 in Gehler''s paper';
                KteProd='Complete the code to continue..\Tip: section 3 in Gehler''s paper';
                
                error(KtrAvg);
                
                Ktrain(:,:,8)=KtrAvg;
                Ktest(:,:,8)=KteAvg;
                
                Ktrain(:,:,9)=KtrProd;
                Ktest(:,:,9)=KteProd;
                
                % generate the labels of the dataset
                Ytrain = zeros(680, 1);
                Ytest  = zeros(340, 1);
                
                for i=1:17,
                    Ytrain((i-1)*40+1:(i-1)*40+40, :) = i;
                    Ytest((i-1)*20+1:(i-1)*20+20, :)  = i;
                end
                save(feat_file, 'Ktrain','Ktest','Ytrain','Ytest','g','-v7.3');
            end
        end
        
        function [Ktrain, Ktest, Ytrain, Ytest]= getCaltech101Kernels(obj,s)
            %COMPUTES THE TRAINING Ktrain AND TESTING Ktest KERNELS
            %
            %PARAMETERS:
            % s is the number of training samples/class. it cam be only 5,
            %15, or 30
            %
            %RETURNS:
            % Ktrain is the 3D training kernel matrix
            % Ktest is the 3D testing kernel matrix
            % Ytrain is the vector of training labels
            % Ytest is the vector of testing labesls
            % g is the vector of parameters of the kernels
            
            labels_file=sprintf('%s/labels/caltech101_nTrain%d_nTest50_N1.mat',obj.datasetPath,s);
            tmp=load(labels_file);
            Ytrain=tmp.tr_label;
            Ytest=tmp.te_label;
            basic_dir=sprintf('%s/n%d_N1',obj.datasetPath,s);
            tr_basic_filename=sprintf('caltech101_nTrain%d_nTest50_N1_feat',s);
            te_basic_filename=sprintf('caltech101_nTrain%d_nTest50_N1_test_feat',s);
            Ktrain=[];
            Ktest=[];
            
            for f=1:7
                tic
                tr_load_file=sprintf('%s/%s%d.mat',basic_dir,tr_basic_filename,f);
                tr_tmp=load(tr_load_file);
                Ktrain(:,:,f)=double(tr_tmp.K);
                
                clear tr_tmp;
                te_load_file=sprintf('%s/%s%d.mat',basic_dir,te_basic_filename,f);
                te_tmp=load(te_load_file);
                Ktest(:,:,f)=double(te_tmp.K);
                clear te_tmp;
            end
            
            %AVERAGE KERNEL
            KtrAvg='Complete the code to continue..\Tip: section 3 in Gehler''s paper';
            KteAvg='Complete the code to continue..\Tip: section 3 in Gehler''s paper';
            
            %PRODUCT KERNEL
            KtrProd='Complete the code to continue..\Tip: section 3 in Gehler''s paper';
            KteProd='Complete the code to continue..\Tip: section 3 in Gehler''s paper';
            
            error(KtrAvg);
            
            Ktrain(:,:,8)=KtrAvg;
            Ktest(:,:,8)=KteAvg;
            
            Ktrain(:,:,9)=KtrProd;
            Ktest(:,:,9)=KteProd;
        end
        
    end
    
    methods (Static)
        
        function [K,gamma]=computeChiSquareKernel(D,gamma)
            % COMPUTES THE CHI-SQUARE KERNEL K, USING THE PRECOMPUTED
            % CHI-SQUARES DISTANCE MATRIX D AND PARAMETER gamma
            %
            % IF gamma IS NOT PASSED AS ARGUMENT, IT CALLS THE tuneGamma
            % METHOD WHICH RETURNS THE APPROPRIATE VALUE FOR THE DISTANCE
            % MATRIX IN USAGE
            
            %%% TUNES THE GAMMA
            if ~exist('gamma','var')
                gamma=MklExperience.tuneGamma(D);
            end
            
            %%% COMPUTES THE KERNEL
            K='Complete the code to continue..\Tip: section 6 in Gehler''s paper';
            error(K);
        end
        
        function gamma=tuneGamma(D)
            %PROVIDED A DISTANCE MATRIX D, IT COMPUTES THE INVERSE OF THE
            %AVERAGE PAIRWISE DISTANCE
            
            gamma='Complete the code to continue..\Tip: section 6 in Gehler''s paper';
            error(gamma);
        end
        
        function model=trainOva(classifier,kernel,labels,C)
            %TRAINS A ONE VS ALL CLASSIFIER
            %
            %PARAMETERS
            %classifier is either 'svm', or 'mkl'
            %kernels is a kernel matrix
            %labels is a vector of labels
            %C is the regularizatione parameter of the classifier
            
            l=unique(labels);
            fprintf('Training using One-VS-All scheme\n');
            for c=1:max(l)
                fprintf('\t class %d\n',c);
                y_tr=(labels==c)*2-1;
                if strcmp(classifier,'svm')
                    model.model(c)=MklExperience.trainSVM(kernel,y_tr,C);
                elseif strcmp(classifier,'mkl')
                    model.model(c)=error('Complete the code to continue..\nCall the trainMKL method with the proper parameters.\Tip: eq 1) in Gehler''s paper');
                else
                    error('classifier must be either ''svm'', or ''mkl''');
                end
            end
        end
        
        function [dec_values,predict_labels,accuracy,confusion,mca]=testOva(classifier,kernel,labels,model)
            %TESTS THE ONE VS ALL CLASSIFIER AND RETURNS: decision
            %values, predicted labels, accuracy, confusion matrix and mean
            %class accuracy
            %
            %PARAMETERS
            %classifier is either 'svm', or 'mkl'
            %kernels is a kernel matrix
            %labels is a vector of labels
            %model is the trained model
            
            nBest=1;
            nCla=numel(model.model);
            nSamp=size(kernel,1);
            fprintf('Testing using One-VS-All scheme\n');
            dec_values=zeros(nSamp,nCla);
            for c=1:nCla
                fprintf('\t class %d\n',c);
                y_te=(labels==c)*2-1;
                if strcmp(classifier,'svm')
                    tmp_dec_values= MklExperience.testSVM(kernel, y_te, model.model(c));
                elseif strcmp(classifier,'mkl')
                    tmp_dec_values= MklExperience.testMKL(kernel, y_te, model.model(c));
                else
                    error('classifier must be either ''svm'', or ''mkl''');
                end
                if isempty(tmp_dec_values)
                    tmp_dec_values=-inf(size(tmp_dec_values,1),1);
                end
                try
                    dec_values(:,c)=tmp_dec_values;
                catch
                    error('cannot process decision values\n');
                end
            end
            [slack,predict_labels_M]=sort(dec_values,2,'descend');
            
            tmp0=predict_labels_M(:,1:nBest);
            tmp=tmp0==repmat(labels,1,nBest);
            [slack,i]=max(tmp,[],2);
            predict_labels=diag((tmp0(1:numel(i),i)));
            
            accuracy=sum(predict_labels==labels)/numel(labels)*100;
            confusion=MklExperience.confmat(uint8(labels),uint8(predict_labels));
            
            mca=error('Complete the code to continue..\nTip: use the confusion matrix to compute the mean class accuracy');
        end
        
        function model=trainSVM(kernel,labels,C)
            %TRAINS AN SVM AND RETURNS THE TRAINED model
            %
            %PARAMETERS:
            %kernel is the training kernel matrix
            %            %
            %labels is the vector with the labels
            %
            %C is the SVM regularization parameter (higher C means less
            %regularization)
            
            n =numel(labels);
            param='-t 4';
            if exist('C','var')
                param=sprintf('%s -c %d',param,C);
            end
            model=svmtrain(double(labels), double([(1:n)', kernel]),param );
            clear kernel labels;
        end
        
        function [dec_values,predict_labels,accuracy]=testSVM(kernel,labels,model)
            %TESTS THE SVM CLASSIFIER AND RETURNS: decision
            %values, predicted labels and accuracy
            %
            %PARAMETERS:
            %kernel is the testing kernel matrix
            %
            %labels is the vector with the labels
            %
            %model is a trained SVM model
            
            n =numel(labels);
            [predict_labels, accuracy, dec_values] = svmpredict(double(labels), double([(1:n)',kernel]), model);
            if numel(model.Label)==2
                dec_values=dec_values*sign(model.Label(1));
            end
            accuracy=accuracy(1);
        end
        
        function model = trainMKL(kernel, labels, C, mkl_p_norm, mkl_eps, svm_eps)
            % Input
            %   kernel: a 3D matrix of M NxN kernel matrices
            %   labels: (N,1) vector of binary labels -1, 1.
            %   C: SVM C scalar.
            %   mkl_p_norm: the norm used to regularize the kernel weights
            %   mkl_eps: (optional) MKL convergence epsilon.
            %   svm_eps: (optional) SVM convergence epsilon.
            %
            % Output
            %   model.w: (M,1) convex MKL weights, sum(model.w) = 1, model.w >= 0.
            %   model.alpha: (N,1) optimal SVM alpha weights.
            %   model.b: scalar, optimal SVM bias.
            %
            % The code is mostly from mkl_classify_christmas_star.m, in the shogun examples/ folder.
            
            if nargin < 5
                mkl_eps = 1e-4;
            end
            if nargin < 6
                svm_eps = 1e-5;
            end
            
            assert(ndims(kernel)==3);
            assert(size(kernel,1)==size(kernel,2));
            
            M = size(kernel,3);
            if M < 1
                error(['You must provide at least one kernel to train the mkl model.']);
            end
            N = size(kernel,1);
            
            % train MKL
            sg('clean_features', 'TRAIN');
            sg('set_labels', 'TRAIN', labels');
            
            sg('new_classifier', 'MKL_CLASSIFICATION');
            
            sg('mkl_use_interleaved_optimization',1);
            sg('set_solver', 'DIRECT');
            sg('mkl_parameters', mkl_eps, 0, mkl_p_norm);
            sg('svm_epsilon',svm_eps);
            
            sg('clean_kernel');
            sg('set_kernel', 'COMBINED', 0);
            
            for m=1:M
                sg('add_kernel',1,'CUSTOM', double(kernel(:,:,m)),'FULL');
            end
            
            sg('c',C);
            
            sg('train_classifier');
            [b, a]=sg('get_svm');
            alpha=zeros(N,1);
            alpha(1+a(:,2))=a(:,1);
            ws=sg('get_subkernel_weights');
            ws=ws';
            sg('clean_features' ,'TRAIN');
            sg('clean_kernel') ;
            
            model.w=ws;
            model.alpha=alpha;
            model.b=b;
        end
        
        function [dec_values,predict_labels,accuracy]=testMKL(kernel,labels,model)
            %TESTS THE SVM CLASSIFIER AND RETURNS: decision
            %values, predicted labels and accuracy
            %
            %PARAMETERS:
            %kernel is the testing kernel matrix
            %
            %labels is the vector with the labels
            %
            %model is a trained MKL model
            
            n =numel(labels);
            dec_values  = Kbeta(kernel,model.w) * model.alpha + model.b;
            predict_labels=sign(dec_values);
            accuracy=(predict_labels==labels)/n;
        end
        
        function c = confmat(label,prediction)
            %COMPUTES THE CONFUSION MATRIX
            %label is the deired output
            %prediction is the predicted output
            
            miny = min(label);
            maxy = max(label);
            c = zeros(maxy-miny);
            for i = miny:maxy
                index = label == i;
                for j = miny:maxy
                    z = prediction(index);
                    c(i-miny+1,j-miny+1) = length(find(z == j));
                end
            end
        end
    end
end
